
// C8051F30x Programmer
// Copyright (C) S.Fuchita, fi7s-fct@asahi-net.or.jp

#include <stdlib.h>
#include <stdio.h>

#ifdef WIN32
	#include <windows.h>
	struct termios {
		int nothing;
	};
#endif
#ifdef LINUX
	#include <unistd.h>
	#include <fcntl.h>
	#include <termios.h>
	#include <sys/stat.h>
	typedef int HANDLE;
#endif

#include "../c2avr/c2.h"

#define TIMEOUT 100000

#define ID_C8051F30X 0x04

////////////////////////////////////////////////////////////////////////

typedef int bool;
enum { false, true };

#define VERSION "2.1"

////////////////////////////////////////////////////////////////////////


#ifdef WIN32
	#define wait_20ms() Sleep(20)
#endif
#ifdef LINUX
	#define wait_20ms() usleep(20000)
#endif

#ifdef LINUX
void noncanonical(HANDLE fd)
{
	struct termios tios;
	tcgetattr(fd, &tios);
	tios.c_lflag &= ~ICANON & ~ECHO;
	tios.c_cc[VTIME] = 0;
	tios.c_cc[VMIN] = 1;
	tcsetattr(fd, TCSANOW, &tios);
}
#endif

// serial port

HANDLE serial_open(const char *device, unsigned long speed, struct termios *saveattr)
{
	HANDLE fd;
#ifdef WIN32
	DCB dcb;
	fd = CreateFile(device, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	GetCommState(fd, &dcb);
	dcb.BaudRate = speed;
	dcb.fBinary = TRUE;
	dcb.fParity = FALSE;
	dcb.fOutxCtsFlow = FALSE;
	dcb.fOutxDsrFlow = FALSE;
	dcb.fDtrControl = DTR_CONTROL_ENABLE;
	dcb.fDsrSensitivity = FALSE;
	dcb.fTXContinueOnXoff = FALSE;
	dcb.fOutX = FALSE;
	dcb.fInX = FALSE;
	dcb.fErrorChar = FALSE;
	dcb.fNull = FALSE;
	dcb.fRtsControl = RTS_CONTROL_ENABLE;
	dcb.fAbortOnError = FALSE;
	dcb.ByteSize = 8;
	dcb.Parity = NOPARITY;
	dcb.StopBits = ONESTOPBIT;
	SetCommState(fd, &dcb);
#endif
#ifdef LINUX
	struct termios attr;

	fd = open(device, O_RDWR | O_NOCTTY | O_NONBLOCK);
	if (fd < 0) {
		fprintf(stderr, "Cannot open: %s\n", device);
		return -1;
	}

	tcgetattr(fd, &attr);
	if (saveattr) *saveattr = attr;

	switch (speed) {
//	case      50:	speed =      B50;		break;
//	case      75:	speed =      B75;		break;
//	case     110:	speed =     B110;		break;
//	case     134:	speed =     B134;		break;
//	case     150:	speed =     B150;		break;
// 	case     200:	speed =     B200;		break;
	case     300:	speed =     B300;		break;
	case     600:	speed =     B600;		break;
	case    1200:	speed =    B1200;		break;
//	case    1800:	speed =    B1800;		break;
	case    2400:	speed =    B2400;		break;
	case    4800:	speed =    B4800;		break;
	case    9600:	speed =    B9600;		break;
	case   19200:	speed =   B19200;		break;
	case   38400:	speed =   B38400;		break;
	case   57600:	speed =   B57600;		break;
	case  115200:	speed =  B115200;		break;
	case  230400:	speed =  B230400;		break;
//	case  460800:	speed =  B460800;		break;
//	case  500000:	speed =  B500000;		break;
//	case  576000:	speed =  B576000;		break;
//	case  921600:	speed =  B921600;		break;
//	case 1000000:	speed = B1000000;		break;
//	case 1152000:	speed = B1152000;		break;
//	case 1500000:	speed = B1500000;		break;
//	case 2000000:	speed = B2000000;		break;
//	case 2500000:	speed = B2500000;		break;
//	case 3000000:	speed = B3000000;		break;
//	case 3500000:	speed = B3500000;		break;
//	case 4000000:	speed = B4000000;		break;
	default:
		speed = B19200;
		break;
	}

	cfsetispeed(&attr, speed);
	cfsetospeed(&attr, speed);
	cfmakeraw(&attr);
	
	attr.c_cflag |= CS8 | CLOCAL | CREAD;
	attr.c_iflag = 0;
	attr.c_oflag = 0;
	attr.c_lflag = 0;
	attr.c_cc[VMIN] = 1;
	attr.c_cc[VTIME] = 0;

	tcsetattr(fd, TCSANOW, &attr);
#endif
	return fd;
}

void serial_close(HANDLE fd, struct termios *attr)
{
#ifdef WIN32
	CloseHandle(fd);
#endif
#ifdef LINUX
	tcsetattr(fd, TCSANOW, attr);
	close(fd);
#endif
}

unsigned int serial_write(HANDLE fd, const void *ptr, unsigned int len)
{
#ifdef WIN32
	WriteFile(fd, ptr, len, &len, NULL);
	return len;
#endif
#ifdef LINUX
	return write(fd, ptr, len);
#endif
}

unsigned int serial_read(HANDLE fd, unsigned char *ptr, unsigned int len, unsigned int timeout)
{
	unsigned long bytes;
#ifdef WIN32
	COMMTIMEOUTS to;
	GetCommTimeouts(fd, &to);
	to.ReadIntervalTimeout = 0;
	to.ReadTotalTimeoutMultiplier = 0;
	to.ReadTotalTimeoutConstant = timeout / 1000;
	to.WriteTotalTimeoutMultiplier = 0;
	to.WriteTotalTimeoutConstant = 0;
	SetCommTimeouts(fd, &to);
	ReadFile(fd, ptr, len, &bytes, NULL);
#endif
#ifdef LINUX
	fd_set fds;
	struct timeval to;
	unsigned long n;
	unsigned char c;

	bytes = 0;
	while (len > 0) {
		FD_ZERO(&fds);
		FD_SET(fd, &fds);
		to.tv_sec = 0;
		to.tv_usec = timeout;
		n = select(fd + 1, &fds, NULL, NULL, &to);
		if (!FD_ISSET(fd, &fds)) break;
		n = read(fd, ptr, len);
		bytes += n;
		len -= n;
		ptr += n;
	}
#endif
	return bytes;
}

int serial_getchar(HANDLE fd, unsigned int timeout)
{
	unsigned char c;
	if (serial_read(fd, &c, 1, timeout) == 1) {
		return c;
	}
	return -1;
}

void serial_flush_input(HANDLE fd, unsigned int timeout)
{
	unsigned char tmp[256];
	while (serial_read(fd, tmp, sizeof(tmp), timeout) > 0);
}

// reset

void c2Reset(HANDLE fd)
{
	unsigned char c;
	c = C2_A_RESET;
	serial_write(fd, &c, 1);
}

// C2 register I/O

void c2WriteAR(HANDLE fd, int addr)
{
	unsigned char tmp[4];
	tmp[0] = 0;
	tmp[1] = ((addr & 0x80) >> 1) | 0x80;
	tmp[2] = addr | 0x80;
	tmp[3] = C2_A_WRITE_ADDRESS;
	serial_write(fd, tmp, 4);
}

int c2ReadAR(HANDLE fd)
{
	unsigned char c;
	c = C2_A_READ_ADDRESS;
	serial_write(fd, &c, 1);
	return serial_getchar(fd, TIMEOUT);
}

void c2WriteDR(HANDLE fd, int data)
{
	unsigned char tmp[4];
	tmp[0] = 0;
	tmp[1] = ((data & 0x80) >> 1) | 0x80;
	tmp[2] = data | 0x80;
	tmp[3] = C2_A_WRITE_DATA;
	serial_write(fd, tmp, 4);
}

int c2ReadDR(HANDLE fd)
{
	unsigned char c;
	c = C2_A_READ_DATA;
	serial_write(fd, &c, 1);
	return serial_getchar(fd, TIMEOUT);
}

//

void c2PollOutReady(HANDLE fd)
{
	int i;
	for (i = 0; !(c2ReadAR(fd) & 1); i++);
}

void c2PollInBusy(HANDLE fd)
{
	int i;
	for (i = 0; c2ReadAR(fd) & 2; i++);
}

void _send_binary(HANDLE fd, unsigned char *ptr, int len)
{
	int i, n;
	unsigned char c, tmp[8];

	c = 0;
	serial_write(fd, &c, 1);

	while (len > 0) {
		n = len;
		if (n > 7) n = 7;
		c = 1;
		for (i = 0; i < 7; i++) {
			c <<= 1;
			if (i < len && (ptr[i] & 0x80)) c++;
		}
		tmp[0] = c;
		for (i = 0; i < n; i++) {
			tmp[i + 1] = *ptr | 0x80;
			ptr++;
			len--;
		}
		serial_write(fd, tmp, n + 1);
	}
}

//

int c2GetDeviceID(HANDLE fd)
{
	c2Reset(fd);
	c2WriteAR(fd, C2_R_DEVICEID);
	return c2ReadDR(fd);
}

int c2GetRevisionID(HANDLE fd)
{
	c2Reset(fd);
	c2WriteAR(fd, C2_R_REVID);
	return c2ReadDR(fd);
}

bool c2BlockRead(HANDLE fd, unsigned char *ptr, unsigned short addr, unsigned char len)
{
	int i, n;
	unsigned char tmp[19];

	while (len > 0) {
		n = len;
		if (n > 16) n = 16;
		tmp[0] = (unsigned char)(addr >> 8);
		tmp[1] = (unsigned char)(addr);
		tmp[2] = n;
		_send_binary(fd, tmp, 3);
		tmp[0] = C2_A_BLOCK_READ;
		tmp[1] = C2_A_TRANSFER;
		serial_write(fd, tmp, 1);
		serial_getchar(fd, TIMEOUT);
		serial_write(fd, tmp + 1, 1);
		i = serial_read(fd, tmp, n + 2, TIMEOUT);
		if (i != n + 2) return false;
		memcpy(ptr, tmp + 1, n);
		addr += n;
		ptr += n;
		len -= n;
	}

	return true;
}

bool c2BlockWrite(HANDLE fd, const unsigned char *ptr, unsigned short addr, unsigned char len)
{
	int i, n;
	unsigned char tmp[19];

	while (len > 0) {
		n = len;
		if (n > 16) n = 16;
		tmp[0] = (unsigned char)(addr >> 8);
		tmp[1] = (unsigned char)(addr);
		tmp[2] = n;
		memcpy(tmp + 3, ptr, n);
		_send_binary(fd, tmp, n + 3);
		tmp[0] = C2_A_BLOCK_WRITE;
		serial_write(fd, tmp, 1);
		i = serial_getchar(fd, TIMEOUT);
		if (i) return false;
		addr += n;
		ptr += n;
		len -= n;
	}

	return true;
}

int c2BlockChecksum(HANDLE fd, unsigned short addr, unsigned char len)
{
	int c, n;
	unsigned char sum, tmp[3];

	sum = 0;

	while (len > 0) {
		n = len;
		if (n > 16) n = 16;
		tmp[0] = (unsigned char)(addr >> 8);
		tmp[1] = (unsigned char)(addr);
		tmp[2] = n;
		_send_binary(fd, tmp, 3);
		tmp[0] = C2_A_BLOCK_READ;
		serial_write(fd, tmp, 1);
		serial_getchar(fd, TIMEOUT);
		tmp[0] = C2_A_CHECKSUM;
		serial_write(fd, tmp, 1);
		c = serial_getchar(fd, TIMEOUT);
		if (c < 0) return -1;
		sum += c;
		addr += n;
		len -= n;
	}

	return sum;
}


#if 0
bool c2FlashProgramMode(HANDLE fd)
{
	c2Reset(fd);
	c2WriteAR(fd, C2_R_FPCTL);
	c2WriteDR(fd, 0x02);
	c2WriteDR(fd, 0x01);
	wait_20ms();
	return true;
}
#else
bool c2FlashProgramMode(HANDLE fd)
{
	int i;
	unsigned char c;

	for (i = 0; i < 10; i++) {
		serial_flush_input(fd, TIMEOUT);
		c = C2_A_FLASH_PROGRAM_MODE;
		serial_write(fd, &c, 1);
		if (serial_getchar(fd, TIMEOUT) == 0) return true;
	}
	return false;
}
#endif

bool c2DeviceErase(HANDLE fd)
{
	unsigned char c;

	c2WriteAR(fd, C2_R_FPDAT);
	c2WriteDR(fd, C2_C_DEVICE_ERASE);
	c2PollInBusy(fd);
	c2PollOutReady(fd);
	if (c2ReadDR(fd) != C2_S_COMMAND_OK)
		return false;

	c2WriteDR(fd, 0xde);
	c2PollInBusy(fd);
	c2WriteDR(fd, 0xad);
	c2PollInBusy(fd);
	c2WriteDR(fd, 0xa5);
	c2PollInBusy(fd);
	c2PollOutReady(fd);

	c = C2_A_ECHO;
	serial_write(fd, &c, 1);
	return serial_getchar(fd, TIMEOUT) == '!';
}

////////////////////////////////////////////////////////////////////////

enum Command {
	C_NONE,
	C_HELP,
	C_READ,
	C_WRITE,
	C_DEVICE_ERASE,
	C_INQUIRE,

	C_TEST,
};


struct CommandInfo {
	int argc;
	char **argv;

	enum Command command;
	bool error;

	const char *serial;
	const char *intelhex;
	unsigned int flashsize;
	unsigned long speed;
	char reset;
};

struct CommandTable {
	const char *name;
	void (*proc)(struct CommandInfo *info, const char *arg);
};

int _sort_command_table(const void *foo, const void *bar)
{
	struct CommandTable *a = (struct CommandTable *)foo;
	struct CommandTable *b = (struct CommandTable *)bar;
	return strcmp(a->name, b->name);
}

void dispatch(const struct CommandTable *table, int count, struct CommandInfo *info)
{
	int c, i, j, k, n, t;
	bool done;

	while (info->argc > 0) {
		char *str = *info->argv++;
		info->argc--;
		done = false;
		if (*str == '-') {
			i = 0;
			n = count;
			while (n > 0) {
				j = i + n / 2;
				k = 0;
				while (1) {
					c = str[k];
					if (c == '=') c = 0;
					t = c - table[j].name[k];
					if (t || !c) break;
					k++;
				}
				if (t < 0) {
					n = j - i;
				} else if (t > 0) {
					n = i + n - j - 1;
					i = j + 1;
				} else {
					switch (str[k]) {
					case 0:		table[j].proc(info, NULL);				break;
					case '=':	table[j].proc(info, str + j + 1);		break;
					}
					done = true;
					break;
				}
			}
			if (!done) {
				fprintf(stderr, "Unknown option: %s\n", str);
				if (info->command == C_HELP) {
					info->command = C_NONE;
				}
			}
		}
	}
}


char *_get_next_arg(struct CommandInfo *info)
{
	if (info->argc > 0) {
		if (**info->argv != '-') {
			char *arg = *info->argv;
			info->argc--;
			info->argv++;
			return arg;
		}
	}
	return NULL;
}

void option_serial(struct CommandInfo *info, const char *arg)
{
	if (!arg) {
		arg = _get_next_arg(info);
		if (!arg) {
			fprintf(stderr, "Serial port is required: --serial <path>\n");
			info->error = true;
			return;
		}
	}
	info->serial = arg;
}

void option_speed(struct CommandInfo *info, const char *arg)
{
	unsigned long speed;

	if (!arg) {
		arg = _get_next_arg(info);
		if (!arg) {
			fprintf(stderr, "Baud rate is required: --speed <bits per second>\n");
			info->error = true;
			return;
		}
	}

	speed = arg ? atol(arg) : 0;

	switch (speed) {
//	case      50:	break;	
//	case      75:	break;	
//	case     110:	break;	
//	case     134:	break;	
//	case     150:	break;	
// 	case     200:	break;	
	case     300:	break;	
	case     600:	break;	
	case    1200:	break;	
//	case    1800:	break;	
	case    2400:	break;	
	case    4800:	break;	
	case    9600:	break;	
	case   19200:	break;	
	case   38400:	break;	
	case   57600:	break;	
	case  115200:	break;	
	case  230400:	break;	
//	case  460800:	break;	
//	case  500000:	break;	
//	case  576000:	break;	
//	case  921600:	break;	
//	case 1000000:	break;	
//	case 1152000:	break;	
//	case 1500000:	break;	
//	case 2000000:	break;	
//	case 2500000:	break;	
//	case 3000000:	break;	
//	case 3500000:	break;	
//	case 4000000:	break;	
	default:
		fprintf(stderr, "Invalid baud rate: --speed <bits per second>\n");
		return;
	}

	info->speed = speed;
}

#define COMMAND "c8051f30"

void command_help(struct CommandInfo *info)
{
	printf("C8051F30x Programmer, v" VERSION "\r\n"
		"\n"
		"commands:\n"
		"    -r --read                           : Read from flash memory\n"
		"    -w --write                          : Write to flash memory\n"
		"       --device-erase                   : Erase flash memory entirely\n"
		"    -i --inquire                        : Inquire target device\n"
		"    -h --help                           : Display this message\n"
		"options:\n"
		"    -f --file <hex file>                : Hex file for serial_write command\n"
#ifdef WIN32
		"       --serial <serial port>           : Serial port (default=COM1)\n"
#endif
#ifdef LINUX
		"       --serial <serial port>           : Serial port (default=/dev/ttyS0)\n"
#endif
		"       --speed <bits per second>        : Baud rate (default=19200)\n"
		"examples:\n"
		"    " COMMAND " -i\n"
		"    " COMMAND " -r\n"
		"    " COMMAND " -w <hexfile\n"
		"    " COMMAND " -w -f hexfile\n"
		"    " COMMAND " --device-erase\n"
#ifdef WIN32
		"    " COMMAND " --serial \\\\.\\COM5 --speed 115200 -i\n"
#endif
#ifdef LINUX
		"    " COMMAND " --serial /dev/ttyS1 --speed 115200 -i\n"
#endif
		"\n"
	);
}
void option_help(struct CommandInfo *info, const char *arg)
{
	info->command = C_HELP;
}


void option_terminal(struct CommandInfo *info, const char *arg)
{
#ifdef LINUX
	int n, fd_tty, fd_ser;
	fd_set fds;
	struct termios tios_tty;
	struct termios ser_attr;

	fd_ser = serial_open(info->serial, info->speed, &ser_attr);
	if (fd_ser < 0) {
		info->error = true;
		return;
	}

	fd_tty = 0;		// stdin
	tcgetattr(fd_tty, &tios_tty);
	noncanonical(fd_tty);

	while (1) {
		FD_ZERO(&fds);
		FD_SET(fd_tty, &fds);		// stdin
		FD_SET(fd_ser, &fds);		// serial
		n = select(fd_ser + 1, &fds, NULL, NULL, NULL);

		if (FD_ISSET(fd_tty, &fds)) {		// transmit
			int c = getchar();
			if (c == '\n') c = '\r';
			if (c >= 0) {
				unsigned char t = c;
				serial_write(fd_ser, &t, 1);
			}
		}

		if (FD_ISSET(fd_ser, &fds)) {	// receive
			unsigned char c;
			if (read(fd_ser, &c, 1) == 1) {
				putchar(c);
				fflush(stdout);
			}
		}
	}

	serial_close(fd_ser, &ser_attr);
#endif
}

unsigned int _get_version(HANDLE fd, char *ptr, unsigned int len)
{
	unsigned char c;
	c = C2_A_GET_VERSION;
	serial_write(fd, &c, 1);
	return serial_read(fd, ptr, len, TIMEOUT);
}

unsigned char _inquire_device(HANDLE fd)
{
	int dev, rev;

	// inquire

	dev = c2GetDeviceID(fd);
	rev = c2GetRevisionID(fd);

	// result

	printf("DeviceID: ");
	printf((dev >= 0 && dev < 0xff) ? "0x%02X" : "FAILED", dev);
	switch (dev) {
	case ID_C8051F30X:
		printf(" (C8051F30x)");
		break;
	}
	putchar('\n');

	printf("RevisionID: ");
	printf((rev >= 0 && rev < 0xff) ? "0x%02X" : "FAILED", rev);
	putchar('\n');

	return dev;
}

bool _do_programming_mode(HANDLE fd)
{
	if (!c2FlashProgramMode(fd)) {
		fprintf(stderr, "Cannot switch to flash programming mode.\n");
		return false;
	}
	return true;
}

void command_inquire(struct CommandInfo *info)
{
	int i, major, minor;
	HANDLE fd_ser;
	unsigned char tmp[100];
	struct termios ser_attr;

	fd_ser = serial_open(info->serial, info->speed, &ser_attr);
	if (fd_ser < 0) {
		info->error = true;
		return;
	}

	printf("Port: %s\n", info->serial);

	// adapter version

	major = -1;
	minor = 0;
	for (i = 0; i < 5; i++) {
		if (_get_version(fd_ser, tmp, sizeof(tmp)) == 4) {
			if (_get_version(fd_ser, tmp + 4, sizeof(tmp) - 4) == 4) {
				if (tmp[0] == 'C' && tmp[1] == '2') {
					if (!memcmp(tmp, tmp + 4, 4)) {
						major = tmp[2];
						minor = tmp[3];
						break;
					}
				}
			}
		}
	}
	if (major < 0) {
		fprintf(stderr, "C2 interface adapter not found.\n");
	} else {
		printf("Adapter: %d.%d\n", major, minor);
		_inquire_device(fd_ser);
	}

	serial_close(fd_ser, &ser_attr);
}

void option_inquire(struct CommandInfo *info, const char *arg)
{
	info->command = C_INQUIRE;
}

void command_device_erase(struct CommandInfo *info)
{
	HANDLE fd_ser;
	struct termios ser_attr;

	fd_ser = serial_open(info->serial, info->speed, &ser_attr);
	if (fd_ser < 0) {
		info->error = true;
		return;
	}

	if (!_do_programming_mode(fd_ser)) {
		info->error = true;
		goto failed;
	}

	if (!c2DeviceErase(fd_ser)) {
		fprintf(stderr, "Device erase failed.\n");
		info->error = true;
		goto failed;
	}

failed:
	serial_close(fd_ser, &ser_attr);
}

void option_device_erase(struct CommandInfo *info, const char *arg)
{
	info->command = C_DEVICE_ERASE;
}

void command_read(struct CommandInfo *info)
{
	int i, n, error;
	HANDLE fd_ser;
	unsigned char sum;
	unsigned char tmp[4 + 16];
	unsigned char hex[1 + (1 + 2 + 1 + 255 + 1) * 2];
	struct termios ser_attr;

	unsigned short pos;
	unsigned short len;

	fd_ser = serial_open(info->serial, info->speed, &ser_attr);
	if (fd_ser < 0) {
		info->error = true;
		return;
	}

	if (!_do_programming_mode(fd_ser)) {
		info->error = true;
		goto failed;
	}

	pos = 0;
	len = info->flashsize;

	error = 0;
	while (len > 0) {
		n = len;
		if (n > 16) n = 16;
		fprintf(stderr, "%04X: reading.\r", pos);
		memset(tmp, 0xff, sizeof(tmp));
		if (c2BlockRead(fd_ser, tmp + 4, pos, n)) {
			for (i = 0; i < n && tmp[i + 4] == 0xff; i++);
			if (i == n) {
				// All data is 0xff, no output
			} else {
				tmp[0] = n;								// data length
				tmp[1] = (unsigned char)(pos >> 8);		// offset hi
				tmp[2] = (unsigned char)(pos);			// offset lo
				tmp[3] = 0;								// record type
				hex[0] = ':';
				sum = 0;
				for (i = 0; i < n + 4; i++) {
					sprintf(hex + i * 2 + 1, "%02X", tmp[i]);
					sum += tmp[i];
				}
				sprintf(hex + i * 2 + 1, "%02X", (unsigned char)-sum);
				puts(hex);
			}
			pos += n;
			len -= n;
		} else {
			error++;
			if (error > 5) {
				break;
			}
			fprintf(stderr, "\r%04X-%04X: read error, retry\n", pos, pos + n - 1);
			c2FlashProgramMode(fd_ser);
		}
	}
	if (error) {
		fprintf(stderr, "Failed.\n");
		info->error = true;
	} else {
		fprintf(stderr, "              \r");
		puts(":00000001FF");
		fprintf(stderr, "Completed.\n");
	}
failed:
	serial_close(fd_ser, &ser_attr);
}

void option_read(struct CommandInfo *info, const char *arg)
{
	info->command = C_READ;
}


bool _write_flash(HANDLE fd, unsigned char *ptr, unsigned short pos, unsigned int len)
{
	int c, i, n, error;
	unsigned char sum;
	unsigned short last_pos;
	unsigned short last_len;

	if (!_do_programming_mode(fd)) 
		return false;

	if (!c2DeviceErase(fd)) {
		fprintf(stderr, "Device erase failed.\n");
		return false;
	}
	fprintf(stderr, "Device erase: ok\n");

	last_pos = pos;
	last_len = len;

	error = 0;
	while (len > 0) {
		n = len;
		if (n > 16) n = 16;

		for (i = 0; i < n && ptr[pos + i] == 0xff; i++);

		if (i == n) {
			pos += n;
			len -= n;
			error = 0;
		} else if (c2BlockWrite(fd, ptr + pos, pos, n)) {
			sum = 0;
			for (i = 0; i < n; i++) {
				sum += ptr[pos + i];
			}
			c = c2BlockChecksum(fd, pos, n);
			if (c == sum) {
				fprintf(stderr, "%04X-%04X: ok\r", pos, pos + n - 1);
				fflush(stdout);
				last_pos = pos;
				last_len = len;
				pos += n;
				len -= n;
				error = 0;
			} else {
				error++;
				fprintf(stderr, "%04X-%04X: checksum error.", pos, pos + n - 1);
			}
		} else {
			error++;
			fprintf(stderr, "%04X-%04X: serial_write error.", pos, pos + n - 1);
		}
		if (error > 0) {
			if (error >= 5) {
				putchar('\n');
				break;
			}
			fprintf(stderr, " retry (%d)\n", error);
			if (!c2FlashProgramMode(fd)) {
				fprintf(stderr, "Unknown error.\n");
				break;
			}
			pos = last_pos;
			len = last_len;
		}
	}
	if (error) {
		fprintf(stderr, "Failed.\n");
	} else {
		fprintf(stderr, "Completed.   \n");
	}
	return !error;
}


int _parse_intel_hex(FILE *fp_hex, void (*dataproc)(const unsigned char *ptr, unsigned short pos, unsigned int len, void *cookie), void *cookie)
{
	int c, i, n, len, status;
	unsigned short offset;
	unsigned char length;
	unsigned char checksum;
	unsigned char tmp[1 + (1 + 2 + 1 + 255 + 1) * 2];

	len = 0;
	status = 0;
	while (1) {
		c = getc(fp_hex);
		if (c < 0 || c == '\n' || c == '\r') {
			if (len > 10 && tmp[0] == ':') {
				for (n = 1; n < len && isxdigit(tmp[n]); n++);
				if (n > 10) {
					n /= 2;
					checksum = 0;
					for (i = 0; i < n; i++) {
						sscanf(tmp + i * 2 + 1, "%02x", &c);
						checksum += c;
						tmp[i] = c;
					}
					length = tmp[0];
					offset = (tmp[1] << 8) | tmp[2];
					if (checksum) {
						status = -1;			// checksum incorrect
					} else {
						switch (tmp[3]) {		// record type
						case 0x00:	// data
							if (n > 5) {
								dataproc(tmp + 4, offset, n - 5, cookie);
							}
							break;
						case 0x01:	// end of file
							status = 1;			// exit
							break;
						case 0x02:	// extended segment address
						case 0x03:	// start segment address
						case 0x04:	// extended linear address
						case 0x05:	// start linear address
							// ignore
							break;
						}
					}
				} else {
					status = -1;				// invalid intel hex format
				}
			}
			len = 0;
		} else if (len < sizeof(tmp)) {
			tmp[len++] = c;
		} else {
			status = -1;					// line too long
		}
		if (status) break;
	}

	return status;
}

void _store_flash_buffer(const unsigned char *ptr, unsigned short pos, unsigned int len, void *cookie)
{
	unsigned char *mem64k = (unsigned char *)cookie;

	while (len > 0) {
		mem64k[pos++ & 0xffff] = *ptr;
		ptr++;
		len--;
	}
}

void command_write(struct CommandInfo *info)
{
	HANDLE fd_ser;
	struct termios ser_attr;
	unsigned char tmp[65536];
	FILE *fp_hex;

	fd_ser = serial_open(info->serial, info->speed, &ser_attr);
	if (fd_ser < 0) {
		info->error = true;
		return;
	}

	fp_hex = stdin;
	if (info->intelhex) {
		fp_hex = fopen(info->intelhex, "ra");
		if (!fp_hex) {
			fprintf(stderr, "Cannot open: %s\n", info->intelhex);
			info->error = true;
			goto failed;
		}
	}

	memset(tmp, 0xff, sizeof(tmp));
	if (_parse_intel_hex(fp_hex, _store_flash_buffer, tmp) < 0) {
		fprintf(stderr, "Invalid file format.\n");
		info->error = true;
		return;
	}
	_write_flash(fd_ser, tmp, 0, info->flashsize);

	if (info->intelhex) fclose(fp_hex);

failed:
	serial_close(fd_ser, &ser_attr);
}

void option_write(struct CommandInfo *info, const char *arg)
{
	info->command = C_WRITE;
}


void option_file(struct CommandInfo *info, const char *arg)
{
	if (!arg) {
		arg = _get_next_arg(info);
		if (!arg) {
			fprintf(stderr, "File name is required: --file <path>\n");
			info->error = true;
			return;
		}
	}
	info->intelhex = arg;
	if (info->command == C_HELP) info->command = C_NONE;
}

void command_reset(struct CommandInfo *info)
{
	HANDLE fd_ser;
	struct termios ser_attr;

	fd_ser = serial_open(info->serial, info->speed, &ser_attr);
	if (fd_ser < 0) {
		info->error = true;
		return;
	}

	c2Reset(fd_ser);

	serial_close(fd_ser, &ser_attr);
}

void option_reset(struct CommandInfo *info, const char *arg)
{
	info->reset = 1;
}

void option_test(struct CommandInfo *info, const char *arg)
{
	info->command = C_TEST;
}

void command_test(struct CommandInfo *info)
{
}





struct CommandTable commandtable[] = {
	"--device-erase",	option_device_erase,
	"--file",			option_file,
	"--help",			option_help,
	"--inquire",		option_inquire,
	"--read",			option_read,
	"--reset",			option_reset,
	"--serial",			option_serial,
	"--speed",			option_speed,
#ifdef LINUX
	"--terminal",		option_terminal,
#endif
	"--test",			option_test,
	"--write",			option_write,
	"-f",				option_file,
	"-h",				option_help,
	"-i",				option_inquire,
	"-r",				option_read,
	"-w",				option_write,
};
static const int commandcount = (sizeof(commandtable) / sizeof(*commandtable));

////////////////////////////////////////////////////////////////////////

int main(int argc, char **argv)
{
	struct CommandInfo info;

	info.command = C_HELP;
	info.error = 0;
#ifdef WIN32
	info.serial = "COM1";
#endif
#ifdef LINUX
	info.serial = "/dev/ttyS0";
#endif
	info.flashsize = 0x1e00;
	info.intelhex = NULL;
	info.speed = 19200;
	info.reset = 0;

	qsort(commandtable, commandcount, sizeof(*commandtable), _sort_command_table);

	info.argc = argc - 1;
	info.argv = argv + 1;
	dispatch(commandtable, commandcount, &info);

	switch (info.command) {
	case C_NONE:			break;
	case C_READ:			command_read(&info);			break;
	case C_WRITE:			command_write(&info);			break;
	case C_DEVICE_ERASE:	command_device_erase(&info);	break;
	case C_INQUIRE:			command_inquire(&info);			break;
	case C_TEST:			command_test(&info);			break;
	default:
		if (info.reset) break;
		if (!info.error) command_help(&info);
		break;
	}
	if (info.reset) command_reset(&info);

	return info.error ? 1 : 0;
}

// end

